/*
 * $Id: log2anon.c,v 1.4 1998/02/09 13:44:09 puhuri Exp puhuri $
 */
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <sys/md5.h>
#include <stdio.h>

#include "hash_db.h"

#include "codes.h"
#include "hierarchy.h"
#include "methodes.h"
#include "protocols.h"

#define SL_TSTAMP  0
#define SL_ELAPSED 1
#define SL_CLIENT  2
#define SL_ACTION  3
#define SL_CODE    4
#define SL_SIZE    5
#define SL_METHOD  6
#define SL_URL     7
#define SL_IDENT   8
#define SL_HIER    9
#define SL_FROM    10
#define SL_CONTENT 11
#define SL_NULL    12
#define TABLESIZE  13

#ifdef __GNUC__
#define __inline inline
#else
#define __inline
#endif
/* 
 * NULL-pointer safe
 */
static __inline void strtolower (char *start, char *stop) {
  char *p;
  if (start != NULL && stop != NULL)
    for (p = start; p != stop; p++)
      *p = tolower(*p);
}
/*
 * finds hostname
 */
static __inline char *findhost (char *s, char **stop) {
  char *p = s;
  if (((p = strchr(p, '/')) != NULL) &&
      *(++p) && *p == '/' && *(++p) &&
      ((*stop = strchr(p, '/')) != NULL || 
       (*stop = strchr(p, ':')) != NULL ||
       (*stop = p + strlen(p))))
    return p;
  else
    *stop = strchr(s, ':');
  return NULL;
}

#define FORMAT_ASCII 1

#define CONTENT_TYPE 1
#define DATLEN 1024
#define DELIM " "
#define ERRMSG __FILE__ ":%d input failed at line %lu\n"

int main(int argc, char **argv)
{
  char input[DATLEN];
  char *tbl[TABLESIZE];
#if defined(FORMAT_SASN)
  utype abuf[DATLEN];
#endif
  unsigned long line = 0;
  double basetime = 0;
  HASH_DB *act, *hiers, *meths, *clis, *servs, *uris, *protos;
#ifdef CONTENT_TYPE
  HASH_DB *cts;
#endif

  act = mkhash(codes, "actions", 509, HASH_TRESH_10p);
  hiers = mkhash(hierarchy, "hierarchies", 257, HASH_TRESH_10p);
  meths = mkhash(methodes, "methods", 31, HASH_TRESH_10p);
  protos = mkhash(protocols, "protocols", 31, HASH_TRESH_10p);
  clis = hash_create("clients", 131071, HASH_TRESH_10p);
  servs = hash_create("servers", 131071, HASH_TRESH_10p);
  uris = hash_create("urils", 262147, HASH_TRESH_50p);
#ifdef CONTENT_TYPE
  cts = hash_create("content-type", 1021, HASH_TRESH_10p);
#endif

#define NONULL(X) if ((X) == NULL) {fprintf(stderr, ERRMSG, __LINE__, line); continue; }
  for(;fgets (input, DATLEN - 1 , stdin) != NULL;line++) {
    double tstamp;
    /*    char *client, *action, *method, *uri, *ident, 
      *hierarchy, *from; */
#ifdef CONTENT_TYPE
    char *content;
    int contentID;
#endif
    int clientID, methodID, hierID, servID, uriID, actionID;
    int protoID = 0;
    int uriHostID = 0;
    unsigned long elapsed, size, ms;
    int code, i;
    char /*p,*/ *s;

    char *last = input + strlen(input) - 1;
    s =  tbl[0] = input;
    if (*last == '\n')
      *last = '\0';		/* remove \n */
    for (i = 1; i < TABLESIZE; i++) {
      if ((s = strchr (s, ' ')) == NULL)
	break;
      *s = '\0';
      while (*(++s) == ' ')	/* eat spaces */
	;
      tbl[i] = s;
      switch (i) {
      case SL_ACTION:
      case SL_HIER:
	if ((s = strchr (s, '/')) != NULL) {
	  *s = '\0';
	  tbl[++i] = ++s;
	} else {
	  fprintf (stderr, __FILE__ 
		   ":%d input failed at line %lu (no / in field %d)\n",
		   __LINE__, line, i);
	}
      }
    }

    if (i == TABLESIZE) {
      fprintf (stderr,
	       __FILE__ ":%d input failed at line %lu (too many fields %d)\n",
	       __LINE__, line, i);
      continue;
    }
    tbl[i] = NULL;

    if ((tstamp = strtod (tbl[SL_TSTAMP], NULL)) == HUGE_VAL) {
      perror ("strtod (tstamp)");
      continue;
    }
    if (basetime == 0.0) {
      basetime = (floor(tstamp / 86400.0)) * 86400.0;
      fprintf (stderr, "basetime: %f\n", basetime);
    }
    ms = (unsigned long)((tstamp - basetime) * 1000);

    if ((elapsed = strtoul (tbl[SL_ELAPSED], NULL, 10)) == ULONG_MAX) {
      perror ("strtoul (elapsed)");
      continue;
    }

    strtolower (tbl[SL_CLIENT], tbl[SL_CLIENT] + strlen(tbl[SL_CLIENT]));
    clientID = hash_put(clis, tbl[SL_CLIENT], strlen(tbl[SL_CLIENT]));

    if ((actionID = hash_get_id (act, tbl[SL_ACTION], strlen(tbl[SL_ACTION])))
	== 0) { 
      fprintf (stderr, "Unknown action: %s\n", tbl[SL_ACTION]);
      actionID = hash_put (act, tbl[SL_ACTION], strlen(tbl[SL_ACTION]));
    }
    if ((code = strtol (tbl[SL_CODE], NULL, 10)) == INT_MAX) {
      perror ("strtoul (code)");
      continue;
    }

    if ((size = strtoul (tbl[SL_SIZE], NULL, 10)) == ULONG_MAX) {
      perror ("strtoul (size)");
      continue;
    }

    if ((methodID = hash_get_id (meths, tbl[SL_METHOD], 
				 strlen(tbl[SL_METHOD]))) == 0) { 
      fprintf (stderr, "Unknown method: %s\n", tbl[SL_METHOD]);
      methodID = hash_put_print (meths, tbl[SL_METHOD], strlen(tbl[SL_METHOD]));
    }

    do {
      char *host, *end = NULL;
      char *s;
      char *proto = tbl[SL_URL];

      if ((host = findhost (tbl[SL_URL], &end)) != NULL || end != NULL)
	strtolower (tbl[SL_URL], end); /* lowercase url start */

      uriID = hash_put(uris, tbl[SL_URL], strlen(tbl[SL_URL]));
      if (host == NULL) {
	protoID = 0;
	if (end == NULL) {
	  uriHostID = hash_put (servs, tbl[SL_URL], strlen(tbl[SL_URL]));
	  break;
	} else {
	  /* Ok, we don't have host and/or protocol */
	  uriHostID = hash_put (servs, tbl[SL_URL], end - tbl[SL_URL] - 1);
	  break;
	}
      }

      /* fprintf (stderr, "URL=%s", proto); */

      if ((s = strchr(proto, ':')) != NULL) {
	int l = s - proto;
	if (l == 4 && strncmp(proto, "http", 4) == 0) {
	  protoID = 1;		/* http */
	} else if (l == 3 && strncmp(proto, "ftp", 3) == 0) {
	  protoID = 2;		/* ftp */
	} else if (l == 6 && strncmp(proto, "gopher", 6) == 0) {
	  protoID = 3;
	} else {
	  *s = '\0';
	  fprintf (stderr, "unknown protocol %s [%d]\n", proto, l);
	  protoID = hash_put_print (protos, proto, l);
	  *s = ':';
	}
	*end = '\0';
	uriHostID = hash_put (servs, host, strlen(host));
      } else {
	protoID = 0;
	uriHostID = 0;
      }
      /* fprintf (stderr, " (proto=%d)\n", protoID);  */
    } while(0);
    if (uriID < 0) {
      fprintf (stderr, "Errors record %lu\n", line);
      continue;
    }

    /* SL_IDENT skipped */

    if ((hierID = hash_get_id (hiers, tbl[SL_HIER], strlen(tbl[SL_HIER]))) 
	== 0) { 
      fprintf (stderr, "Unknown hierarchy: %s\n", tbl[SL_HIER]);
       methodID = hash_put_print (hiers, tbl[SL_HIER], strlen(tbl[SL_HIER]));
    }
    strtolower (tbl[SL_FROM], tbl[SL_FROM] + strlen(tbl[SL_FROM]));
    servID = hash_put (servs, tbl[SL_FROM], strlen(tbl[SL_FROM]));
#ifdef CONTENT_TYPE
    if (tbl[SL_CONTENT] == NULL) {
      contentID = 0; content = "-";
    } else {
      strtolower (tbl[SL_CONTENT], tbl[SL_CONTENT] + strlen(tbl[SL_CONTENT]));
      contentID = hash_put_print (cts, tbl[SL_CONTENT], 
				  strlen(tbl[SL_CONTENT]));
    }
#endif

#if defined(FORMAT_SASN)
    fwrite (abuf, sizeof(utype), sasn_ul(abuf, ms), stdout);
    fwrite (abuf, sizeof(utype), sasn_ul(abuf, elapsed), stdout);
    fwrite (abuf, sizeof(utype), sasn_ul(abuf, size), stdout);
    fwrite (abuf, sizeof(utype), sasn_i(abuf, clientID), stdout);
    fwrite (abuf, sizeof(utype), sasn_i(abuf, uriID), stdout);
    fwrite (abuf, sizeof(utype), sasn_i(abuf, servID), stdout);
    fwrite (abuf, sizeof(utype), 
	    sasn_ul(abuf, 
		    (((((code << 4) |	/* method 7 < 2^4 */
			methodID) << 6) | /* action 34 < 2^6 */
		      actionID) << 5) |	/* hierrachy 20 < 2^5 */
		    hierID),	/* 5 + 6 + 4 + 10 = 25 < 32 */
	    stdout);
#elif defined(FORMAT_ASCII) && defined(CONTENT_TYPE)
    printf ("%lu.%03lu\t%lu\t%lu\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
	    ms/1000, ms%1000, elapsed, size, clientID, uriID, protoID,
	    uriHostID, servID, code, methodID, actionID, hierID, contentID);
#elif defined(FORMAT_ASCII)
    printf ("%lu.%03lu\t%lu\t%lu\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
	    ms/1000, ms%1000, elapsed, size, clientID, uriID, protoID,
	    uriHostID, servID, code, methodID, actionID, hierID);
#elif defined(FORMAT_VERBOSE) && defined(CONTENT_TYPE)
    printf ("Now: %.3f\n\t%s asked %s\n\tGot it in %lu seconds from %s\n\t"
	    "with hier %s, action %s, code %d, content %s\n", 
	    tstamp, client, uri, elapsed, from, hierarchy, action, code,
	    content);
#elif defined(FORMAT_VERBOSE)
    printf ("Now: %.3f\n\t%s asked %s\n\tGot it in %lu seconds from %s\n\t"
	    "with hier %s, action %s, code %d\n", tstamp, client, uri,
	    elapsed, from, hierarchy, action, code);
#elif defined(FORMAT_FIELDS)
    printf ("t=%.3f(+%lu), b=%lu, c=%d, u=%d, s=%d, h=%d, a=%d, C=%d\n",
	    tstamp, elapsed, size, clientID, uriID, servID, hierID,
	    actionID, code);
#else
#error Some FORMAT_* must define
#endif
  }
  hash_stat(act);
  hash_stat(hiers);
  hash_stat(meths);
  hash_stat(clis);
  hash_stat(servs);
  hash_stat(uris);
#ifdef CONTENT_TYPE
  hash_stat(cts);
#endif
  return 0;
}
